import os
import re

class codeCompile:
    '''生成代码类'''
    INDENT_STEP = 4     # 缩进的空格数

    def __init__(self, indent=0):
        self.indent = indent    # 当前缩进
        self.lines = []         # 保存一行一行生成的代码

    def add_indent(self):
        self.indent += self.INDENT_STEP

    def del_indent(self):
        self.indent -= self.INDENT_STEP

    def add(self, code):
        self.lines.append(code)

    def add_line(self, code):
        self.lines.append(' ' * self.indent + code)

    def __str__(self):
        return '\n'.join(map(str, self.lines))

    def __repr__(self):
        return str(self)


class template:
    '''模板转换类'''

    def __init__(self, raw_text, indent=0, default_context=None,
                 func_name='__func_name', result_var='__result',
                 template_dir='', encoding='utf-8'):
        self.raw_text = raw_text
        self.default_context = default_context or {}
        self.func_name = func_name
        self.result_var = result_var
        self.template_dir = template_dir
        self.encoding = encoding
        self.code_builder = code_builder = codeCompile(indent=indent)
        self.buffered = []

        # 变量规则
        self.rule_var = re.compile(r'\{\{ .*? \}\}')
        # 注释规则
        self.rule_comment = re.compile(r'\{# .*? #\}')
        # 标签规则
        self.rule_tag = re.compile(r'\{% .*? %\}')
        # 分割模版字符串
        self.rule_tokens = re.compile(r'''(
            (?:\{\{ .*? \}\})
            |(?:\{\# .*? \#\})
            |(?:\{% .*? %\})
        )''', re.VERBOSE)
        # extends
        self.rule_extends = re.compile(r'\{% extends (?P<name>.*?) %\}')
        # blocks
        self.rule_blocks = re.compile(
            r'\{% block (?P<name>\w+) %\}'
            r'(?P<code>.*?)'
            r'\{% endblock \1 %\}', re.DOTALL)
        # block.super
        self.rule_block_super = re.compile(r'\{\{ block\.super \}\}')

        # 生成 def __func_name():
        code_builder.add_line('def {}():'.format(self.func_name))
        code_builder.add_indent()
        # 生成 __result = []
        code_builder.add_line('{} = []'.format(self.result_var))
        # 解析模版
        self._parse_text()

        self.flush_buffer()
        # 生成 return "".join(__result)
        code_builder.add_line('return "".join({})'.format(self.result_var))
        code_builder.del_indent()

    def _parse_text(self):
        """解析模版"""
        # extends
        self._handle_extends()

        tokens = self.rule_tokens.split(self.raw_text)
        handlers = (
            (self.rule_var.match, self._handle_variable),   # {{ variable }}
            (self.rule_tag.match, self._handle_tag),             # {% tag %}
            (self.rule_comment.match, self._handle_comment),     # {# comment #}
        )
        default_handler = self._handle_string                  # 普通字符串

        for token in tokens:
            for match, handler in handlers:
                if match(token):
                    handler(token)
                    break
            else:
                default_handler(token)

    def _handle_variable(self, token):
        """处理变量"""
        variable = token.strip('{} ')
        self.buffered.append('str({})'.format(variable))

    def _handle_comment(self, token):
        """处理注释"""
        pass

    def _handle_string(self, token):
        """处理字符串"""
        self.buffered.append('{}'.format(repr(token)))

    def _handle_tag(self, token):
        """处理标签"""
        # 将前面解析的字符串，变量写入到 code_builder 中
        # 因为标签生成的代码需要新起一行
        self.flush_buffer()
        tag = token.strip('{%} ')
        tag_name = tag.split()[0]
        if tag_name == 'include':
            self._handle_include(tag)
        else:
            self._handle_statement(tag)

    def _handle_statement(self, tag):
        """处理 if/for"""
        tag_name = tag.split()[0]
        if tag_name in ('if', 'elif', 'else', 'for'):
            # elif 和 else 之前需要向后缩进一步
            if tag_name in ('elif', 'else'):
                self.code_builder.del_indent()
            # if True:, elif True:, else:, for xx in yy:
            self.code_builder.add_line('{}:'.format(tag))
            # if/for 表达式部分结束，向前缩进一步，为下一行做准备
            self.code_builder.add_indent()
        elif tag_name in ('break',):
            self.code_builder.add_line(tag)
        elif tag_name in ('endif', 'endfor'):
            # if/for 结束，向后缩进一步
            self.code_builder.del_indent()

    def _handle_include(self, tag):
        filename = tag.split()[1].strip('"\'')
        included_template = self._parse_another_template_file(filename)

        self.code_builder.add(included_template.code_builder)
        self.code_builder.add_line(
            '{0}.append({1}())'.format(
                self.result_var, included_template.func_name
            )
        )

    def _parse_another_template_file(self, filename):
        template_path = os.path.realpath(
            os.path.join(self.template_dir, filename)
        )
        name_suffix = str(hash(template_path)).replace('-', '_')
        func_name = '{}_{}'.format(self.func_name, name_suffix)
        result_var = '{}_{}'.format(self.result_var, name_suffix)
        with open(template_path, encoding=self.encoding) as fp:
            template = self.__class__(
                fp.read(), indent=self.code_builder.indent,
                default_context=self.default_context,
                func_name=func_name, result_var=result_var,
                template_dir=self.template_dir
            )
        return template

    def _handle_extends(self):
        match_extends = self.rule_extends.match(self.raw_text)
        if match_extends is None:
            return

        parent_template_name = match_extends.group('name').strip('"\' ')
        parent_template_path = os.path.join(
            self.template_dir, parent_template_name
        )
        # 获取当前模版里的所有 blocks
        child_blocks = self._get_all_blocks(self.raw_text)
        # 用这些 blocks 替换掉父模版里的同名 blocks
        with open(parent_template_path, encoding=self.encoding) as fp:
            parent_text = fp.read()
        new_parent_text = self._replace_parent_blocks(
            parent_text, child_blocks
        )
        # 改为解析替换后的父模版内容
        self.raw_text = new_parent_text

    def _replace_parent_blocks(self, parent_text, child_blocks):
        """用子模版的 blocks 替换掉父模版里的同名 blocks"""
        def replace(match):
            name = match.group('name')
            parent_code = match.group('code')
            child_code = child_blocks.get(name, '')
            child_code = self.rule_block_super.sub(parent_code, child_code)
            new_code = child_code or parent_code
            return new_code
        return self.rule_blocks.sub(replace, parent_text)

    def _get_all_blocks(self, text):
        """获取模版内定义的 blocks"""
        return {
            name: code
            for name, code in self.rule_blocks.findall(text)
        }

    def flush_buffer(self):
        # 生成类似代码: __result.extend(['<h1>', name, '</h1>'])
        line = '{0}.extend([{1}])'.format(
            self.result_var, ','.join(self.buffered)
        )
        self.code_builder.add_line(line)
        self.buffered = []

    def render(self, context=None):
        """模版渲染"""
        namespace = {}
        namespace.update(self.default_context)
        if context:
            namespace.update(context)
        exec(str(self.code_builder), namespace)
        result = namespace[self.func_name]()
        return result

def render_template(url,data):
    '''方法封装'''
    fobj=open(url,"r")
    content=fobj.read()
    fobj.close()
    tpl=template(content)
    return tpl.render(data);



import socket
import threading
import re
import os.path
from urllib.parse import urlparse
from .tm import render_template
def demo():
    print("demo")
class server():
    def __init__(self):
        self.host = ""
        self.port = 8888
        self.listen = 512
        self.recvNum=1024
        self.staticDir="static"
        self.ico="1.ico"
        self.staticType=".jpg,.png,.gif,.css,.js,.html"
        self.mimeType={".jpg":"image/jpeg",".png":"image/png",".gif":"image/gif",".css":"text/css",".js":"application/x-javascript",".html":"text/html"}
        self.routeInfo=[]
        self.template_dir="template"
    def route(self,url,methods=["GET"]):
        '''采集路由的信息'''
        def fn(callback):
            obj={}
            obj["params"]=re.findall(r':([^/]+)',url)
            obj["url"]=re.sub(r':[^/]+',"(\w+)",url)
            obj["callback"]=callback
            obj["methods"]=methods
            self.routeInfo.append(obj)
        return fn
    def parse_headers(self,response):
        info=""
        info+=response["headers"]["pool"]+" "+response["headers"]["status_code"]+"\r\n"
        info+="content-type:"+response["headers"]["content-type"]+"\r\n\r\n"

        return info.encode(encoding="utf8")

    def route_handle(self,requestData,clientTcp):
        '''请求的动态的路由做处理'''
        flag=True
        for item in self.routeInfo:
            reg=item["url"]
            result=re.match(r'%s'%(reg),requestData["path"])
            if  result and item["methods"].count(requestData["type"])>0:
                for item1 in range(len(item["params"])):
                    requestData["params"][item["params"][item1]]=result.group(item1+1)

                flag=False
                response={}
                response["headers"]={}
                response["headers"]["pool"]="HTTP/1.1"
                response["headers"]["status_code"]="200 ok"
                response["headers"]["content-type"]="text/html;charset=utf-8"
                body=item["callback"](requestData,response)
                headers=self.parse_headers(response)
                clientTcp.send(headers+body.encode(encoding="utf8"))
                clientTcp.close()
                break
        if flag:

            clientTcp.send(b"HTTP/1.0 404 not find\r\n\r\n this page not find")
            clientTcp.close()

    def requestData_handle(self,requestData):
        '''对请求的数据进行格式化'''
        obj={}
        requestInfo = (requestData.decode(encoding="utf8").splitlines())
        if len(requestInfo)>0:
            firstInfo = requestInfo[0]
            type=re.match(r'[^\s]+',firstInfo).group(0)
            obj["type"]=type
            requestUrl = urlparse(re.match(r'\w+\s+(/[^\s]*)', firstInfo).group(1))

            obj["path"]=requestUrl[2]
            obj["query"]={}
            obj["params"]={}
            if requestUrl[4]:
                for item in requestUrl[4].split("&"):
                    arr=item.split("=")
                    obj["query"][arr[0]]=arr[1]


        return obj

    def request_handle(self,requestData,clientTcp):
        '''处理请求的逻辑'''
        requestUrl=requestData["path"]
        if requestUrl=="/favicon.ico":
            try:
                fobj=open(self.ico,"rb")
                content=fobj.read()
                fobj.close()
                clientTcp.send(b"HTTP/1.1 200 OK\r\ncontent-type:image/x-icon\r\n\r\n"+content)
                clientTcp.close()
            except:
                clientTcp.send(b"HTTP/1.1 404 not find")
                clientTcp.close()
        elif self.staticType.find(os.path.splitext(requestUrl)[1])>-1 and os.path.splitext(requestUrl)[1]:
            if os.path.isdir(self.staticDir):

                fullpath=os.path.join(self.staticDir,requestUrl[1:])

                if os.path.isfile(fullpath):
                    fobj=open(fullpath,"rb")
                    con=fobj.read()
                    fobj.close()
                    clientTcp.send(("HTTP/1.1 200 OK\r\ncontent-type:"+self.mimeType[os.path.splitext(requestUrl)[1]]+";charset=utf-8\r\n\r\n").encode(encoding="utf8")+con)
                    clientTcp.close()

                else:
                    clientTcp.send(b"HTTP/1.1 404 not find")
                    clientTcp.close()
        else:
            # 用户发起请求
            self.route_handle(requestData,clientTcp)

    def client_handle(self,clientTcp,clientInfo):
        '''多线程处理客户端的请求'''
        requestData=clientTcp.recv( self.recvNum)

        # 对每一个请求做处理的的时候，获得请求的信息，并且格式化
        # 实现具体的http协议

        self.request_handle(self.requestData_handle(requestData),clientTcp)

    def createTcp(self):
        '''创建tcp服务器'''
        socketTcp=socket.socket(socket.AF_INET,socket.SOCK_STREAM)
        socketTcp.setsockopt(socket.SOL_SOCKET,socket.SO_REUSEADDR,1)
        socketTcp.bind((self.host,self.port))
        socketTcp.listen(self.listen)
        while True:
            clientTcp,clientInfo=socketTcp.accept()
            threading.Thread(target=self.client_handle,args=(clientTcp,clientInfo)).start()

    def start(self,host="",port=8888,listen=512):
        '''启动函数'''
        self.host=host
        self.port=port
        self.listen=listen
        self.createTcp()





